__all__ = [
    "DrsInvalidScheme",
    "DrsRecordNotFound",
    "DrsRequestError",
]


class DrsInvalidScheme(Exception):
    pass


class DrsRecordNotFound(Exception):
    pass


class DrsRequestError(Exception):
    pass



from logging import Logger
from structlog.stdlib import BoundLogger
from typing import Literal

__all__ = [
    "LogLevelLiteral",
    "StdOrBoundLogger",
]


LogLevelLiteral = Literal["debug", "info", "warning", "error"]
StdOrBoundLogger = Logger | BoundLogger



import structlog
import time
from fastapi import Request, Response
from uvicorn.protocols.utils import get_path_with_query_string

__all__ = [
    "build_structlog_fastapi_middleware",
]

# For use in FastAPI, also see /docs/logging/fastapi_structlog.md


def build_structlog_fastapi_middleware(service_kind: str):
    """
    Helper to build an access-logging middleware for a FastAPI service, formatting access log messages using structlog.
    :param service_kind: Bento Service Kind (e.g., "reference") for the service installing this middleware.
    :return: Access log middleware function for use in a FastAPI instance.
    """

    access_logger = structlog.stdlib.get_logger(f"{service_kind}.access")
    service_logger = structlog.stdlib.get_logger(f"{service_kind}.logger")

    async def access_log_middleware(request: Request, call_next) -> Response:
        start_time = time.perf_counter_ns()

        # To return if an exception occurs while calling the next in the middleware chain
        response: Response = Response(status_code=500)

        try:
            response = await call_next(request)
        except Exception as e:  # pragma: no cover
            await service_logger.aexception("uncaught exception", exc_info=e)
        finally:
            # When the response has finished or errored out, write the access log message:

            duration = time.perf_counter_ns() - start_time

            status_code = response.status_code
            # noinspection PyTypeChecker
            url = get_path_with_query_string(request.scope)
            client_host = request.client.host
            client_port = request.client.port
            http_method = request.method
            http_version = request.scope["http_version"]

            await access_logger.ainfo(
                # The message format mirrors the original uvicorn access message, which we aim to replace here with
                # something more structured.
                f"{client_host}:{client_port} - \"{http_method} {url} HTTP/{http_version}\" {status_code}",
                # HTTP information, extracted from the request and response objects:
                http={
                    "url": url,
                    "status_code": status_code,
                    "method": http_method,
                    "version": http_version,
                },
                # Network information, extracted from the request object:
                network={"client": {"host": client_host, "port": client_port}},
                # Duration in nanoseconds, computed in-middleware:
                duration=duration,
            )

            return response

    return access_log_middleware



from typing import Literal

__all__ = [
    "StreamingException",
    "RangeNotSatisfiableReason",
    "StreamingRangeNotSatisfiable",
    "StreamingBadRange",
    "StreamingProxyingError",
    "StreamingResponseExceededLimit",
    "StreamingBadURI",
    "StreamingUnsupportedURIScheme",
]


class StreamingException(Exception):
    """
    Generic streaming exception / base class for other bento_lib.streaming exceptions.
    """
    pass


RangeNotSatisfiableReason = Literal["start>=length", "end>=length", "inverted", "overlap", "proxied"]


class StreamingRangeNotSatisfiable(StreamingException):
    def __init__(self, message: str, reason: RangeNotSatisfiableReason, n_bytes: int | None):
        self._n_bytes: int | None = n_bytes
        self._reason: RangeNotSatisfiableReason = reason
        super().__init__(message)

    @property
    def reason(self) -> RangeNotSatisfiableReason:
        return self._reason

    @property
    def n_bytes(self) -> int:
        return self._n_bytes


class StreamingBadRange(StreamingException):
    pass


class StreamingProxyingError(StreamingException):
    pass


class StreamingResponseExceededLimit(StreamingException):
    pass


class StreamingBadURI(StreamingException):
    pass


class StreamingUnsupportedURIScheme(StreamingException):
    pass



import re

from .exceptions import StreamingRangeNotSatisfiable, StreamingBadRange

__all__ = ["validate_interval", "parse_range_header"]


BYTE_RANGE_INTERVAL_SPLIT = re.compile(r",\s*")
BYTE_RANGE_START_ONLY = re.compile(r"^(\d+)-$")
BYTE_RANGE_START_END = re.compile(r"^(\d+)-(\d+)$")
BYTE_RANGE_SUFFIX = re.compile(r"^-(\d+)$")


def validate_interval(
    interval: tuple[int, int],
    content_length: int,
    refget_mode: bool = False,
    enforce_not_inverted: bool | None = None,
) -> None:
    """
    Validates a byte interval on a particular content length. Raises an exception (a subclass of
    bento_lib.streaming.exceptions.StreamingException) if the interval is invalid. Raising for inverted intervals is
    optional but on by default unless Refget mode is turned on, in which case it is off unless otherwise set.
    :param interval: A 0-based, inclusive byte interval to validate: a tuple of start byte index, end byte index.
    :param content_length: The total content length of the bytes the interval is for.
    :param refget_mode: Whether to generate Refget-compatible errors rather than standard errors for ranges which go
           past the end of the content bytes.
    :param enforce_not_inverted: Whether to enforce interval order (i.e., start <= end). If None, falls back to
           enforcing only when not in Refget mode.
    :return: None. An exception is raised if the interval is invalid (see bento_lib.streaming.exceptions).
    """

    int_start, int_end = interval

    # Order of these checks is important - we want to give a 416 if start/end is beyond content length (which also
    # results in an inverted interval)

    if int_start >= content_length:
        # both ends of the range are 0-indexed, inclusive - so it starts at 0 and ends at content_length - 1
        if refget_mode:  # sigh... GA4GH moment
            raise StreamingBadRange(f"start is beyond content length: {int_start} >= {content_length}")
        raise StreamingRangeNotSatisfiable(
            f"not satisfiable: {int_start} >= {content_length}", "start>=length", content_length)

    if int_end >= content_length:
        # both ends of the range are 0-indexed, inclusive - so it starts at 0 and ends at content_length - 1
        if refget_mode:  # sigh... GA4GH moment
            raise StreamingBadRange(f"end is beyond content length: {int_end} >= {content_length}")
        raise StreamingRangeNotSatisfiable(
            f"not satisfiable: {int_end} >= {content_length}", "end>=length", content_length)

    should_check_inverted = (not refget_mode) if enforce_not_inverted is None else enforce_not_inverted
    if should_check_inverted and int_start > int_end:
        raise StreamingRangeNotSatisfiable(
            f"inverted interval: {interval}", "inverted", content_length)


def parse_range_header(
    range_header: str | None, content_length: int, refget_mode: bool = False
) -> tuple[tuple[int, int], ...]:
    """
    Parse a range header (given a particular content length) into a validated series of sorted, non-overlapping
    start/end-inclusive intervals.
    """

    if range_header is None:
        return ((0, content_length - 1),)

    intervals: list[tuple[int, int]] = []

    if not range_header.startswith("bytes="):
        raise StreamingBadRange("only bytes range headers are supported")

    intervals_str = range_header.removeprefix("bytes=")

    # Cases: start- | start-end | -suffix, [start- | start-end | -suffix], ...

    intervals_str_split = BYTE_RANGE_INTERVAL_SPLIT.split(intervals_str)

    for iv in intervals_str_split:
        if m := BYTE_RANGE_START_ONLY.match(iv):
            intervals.append((int(m.group(1)), content_length - 1))
        elif m := BYTE_RANGE_START_END.match(iv):
            intervals.append((int(m.group(1)), int(m.group(2))))
        elif m := BYTE_RANGE_SUFFIX.match(iv):
            inclusive_content_length = content_length - 1
            suffix_length = int(m.group(1))  # suffix: -500 === last 500:
            intervals.append((max(inclusive_content_length - suffix_length + 1, 0), inclusive_content_length))
        else:
            raise StreamingBadRange("byte range did not match any pattern")

    intervals.sort()
    n_intervals: int = len(intervals)

    # validate intervals are not inverted and do not overlap each other:
    for i, int1 in enumerate(intervals):
        validate_interval(int1, content_length, refget_mode=refget_mode)

        _, int1_end = int1
        if i < n_intervals - 1:
            int2 = intervals[i + 1]
            int2_start, int2_end = int2

            if int1_end >= int2_start:
                raise StreamingRangeNotSatisfiable(f"intervals overlap: {int1}, {int2}", "overlap", content_length)

    return tuple(intervals)



from datetime import datetime

from .utils import decode_drs_uri, fetch_drs_record_by_uri, fetch_drs_record_by_uri_async

__all__ = [
    "DrsResolver",
]


class DrsResolver:
    def __init__(self, cache_ttl: float = 900.0):
        """
        Constructor for a DrsResolver object, which is used to fetch and cache DRS records by their drs://-schemed URI.
        :param cache_ttl: How long (in seconds) a cache entry is valid for.
        """
        self._cache_ttl: float = cache_ttl
        self._drs_record_cache: dict[str, tuple[float, dict]] = {}

    @staticmethod
    def decode_drs_uri(drs_uri: str) -> str:
        """
        This is just a wrapper for bento_lib.drs.utils.decode_drs_uri.
        :param drs_uri: The drs://-schemed URI for the DRS object.
        :return: The HTTP URI for the DRS object specified.
        """
        return decode_drs_uri(drs_uri)

    def fetch_drs_record_by_uri(self, drs_uri: str) -> dict:
        """
        Fetches and, for a time, caches a DRS record using its drs://-schemed URI.
        Cache is shared between this function and its asynchronous version.
        :param drs_uri: A resolvable drs://-schemed URI for a DRS object record.
        :return: The fetched record dictionary.
        """

        now = datetime.now().timestamp()

        cache_record = self._drs_record_cache.get(drs_uri)
        if cache_record is not None and now - cache_record[0] <= self._cache_ttl:
            return cache_record[1]

        res = fetch_drs_record_by_uri(drs_uri)
        self._drs_record_cache[drs_uri] = (now, res)
        return res

    async def fetch_drs_record_by_uri_async(self, drs_uri: str, session_kwargs: dict | None = None) -> dict:
        """
        Asynchronously fetches and, for a time, caches a DRS record using its drs://-schemed URI.
        Cache is shared between this function and its synchronous version.
        :param drs_uri: A resolvable drs://-schemed URI for a DRS object record.
        :param session_kwargs: Optional dictionary of parameters to pass to the aiohttp.ClientSession constructor.
        :return: The fetched record dictionary.
        """

        now = datetime.now().timestamp()

        cache_record = self._drs_record_cache.get(drs_uri)
        if cache_record is not None and now - cache_record[0] <= self._cache_ttl:
            return cache_record[1]

        res = await fetch_drs_record_by_uri_async(drs_uri, session_kwargs)
        self._drs_record_cache[drs_uri] = (now, res)
        return res



import logging
import structlog.processors
import sys
from structlog.types import EventDict, Processor

from bento_lib.config.pydantic import BentoBaseConfig
from .. import log_level_from_str, LogLevelLiteral


__all__ = [
    "drop_color_message_key",
    "STRUCTLOG_COMMON_PROCESSORS",
    "JSON_LOG_PROCESSORS",
    "CONSOLE_LOG_PROCESSORS",
    "configure_structlog",
    "configure_structlog_from_bento_config",
    "configure_structlog_uvicorn",
]


def drop_color_message_key(_logger, _method_name, event_dict: EventDict) -> EventDict:
    """
    By default, uvicorn includes a color_message key in addition to the log event message, which is formatted for
    coloured terminal output. This processor removes that key if it exists, since it's extraneous in a structured
    logging context.
    """
    event_dict.pop("color_message", None)
    return event_dict


# These common processors are used for all log messages, both ones emitted using structlog and ones formatted from
# standard-library logger objects.
# Most of these are self-explanatory; see https://www.structlog.org/en/stable/processors.html for general information on
# structlog processors.
STRUCTLOG_COMMON_PROCESSORS: list[Processor] = [
    # see https://www.structlog.org/en/stable/api.html#structlog.contextvars.merge_contextvars
    structlog.contextvars.merge_contextvars,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    # format events (messages) with % using positional arguments, like Python's standard logging library:
    structlog.stdlib.PositionalArgumentsFormatter(),
    # add any extra arguments passed as args/kwargs to the log call to the final object:
    structlog.stdlib.ExtraAdder(),
    drop_color_message_key,  # removes uvicorn's "color_message" extra
    structlog.processors.TimeStamper(fmt="iso"),
]

# If we're outputting logs in JSON format, make sure to:
#  - format tracebacks as dictionaries (to render as JSON), and
#  - render the log object as JSON
JSON_LOG_PROCESSORS: list[Processor] = [structlog.processors.dict_tracebacks, structlog.processors.JSONRenderer()]
# If we're outputting 'pretty' logs to stdout, use a normal console renderer
CONSOLE_LOG_PROCESSORS: list[Processor] = [
    structlog.dev.ConsoleRenderer(
        # Use a rich exception formatter, but don't show locals since it ends up being exceedingly verbose:
        exception_formatter=structlog.dev.RichTracebackFormatter(show_locals=False)
    )
]


def _build_root_logger_handler(json_logs: bool) -> logging.StreamHandler:
    """
    Create a stdout stream handler for the root logger, which is responsible for formatting structlog-emitted log
    messages, as well as processing _and_ formatting foreign (stdlib logger-emitted) messages.
    :param json_logs: Whether the root logger stream handler should output messages as JSON or human-readable text.
    """

    # handler to used for every log message
    #  - use stdout instead of stderr: https://12factor.net/logs
    handler = logging.StreamHandler(stream=sys.stdout)
    #  - formatter uses structlog for consistent output between Python logging and structlog logger objects
    #     - foreign_pre_chain is responsible for reformatting external (Python logger) log events, so we have a
    #       consistent starting point between logging and structlog event formats.
    handler.setFormatter(
        structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=STRUCTLOG_COMMON_PROCESSORS,
            processors=[
                # Remove _record & _from_structlog.
                structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                *(JSON_LOG_PROCESSORS if json_logs else CONSOLE_LOG_PROCESSORS),
            ],
        )
    )

    return handler


def configure_structlog(json_logs: bool, log_level: LogLevelLiteral):  # pragma: no cover
    """
    Configure root structlog for the downstream service.
    :param json_logs: Whether to output logs as JSON. If this is false, a human-readable text format is used instead.
    :param log_level: The log level to output at.
    """

    structlog.configure(
        processors=STRUCTLOG_COMMON_PROCESSORS + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    # Set up root logger with a structlog handler and the specified log level.
    root_logger = logging.getLogger()
    root_logger.handlers.clear()
    root_logger.addHandler(_build_root_logger_handler(json_logs))
    root_logger.setLevel(log_level_from_str(log_level))


def configure_structlog_from_bento_config(config: BentoBaseConfig):  # pragma: no cover
    """
    Configure root structlog for the downstream Bento service if the service is using the BentoBaseConfig Pydantic model
    for configuration. The JSON log boolean and output log level are extracted from the config object.
    :param config: BentoBaseConfig-inheriting configuration object instance.
    """
    configure_structlog(json_logs=config.bento_json_logs, log_level=config.log_level)


def configure_structlog_uvicorn():  # pragma: no cover
    """
    Configure uvicorn loggers to be compatible with structlog as configured above. Error/message loggers are changed
    to propogate to the root logger configured in configure_structlog(...), and the default uvicorn access logger is
    *silenced* (to be replaced by a custom access logger elsewhere!)
    """

    for lgr in ("uvicorn", "uvicorn.error"):
        # Propogate these loggers' messages to the root logger (using the StreamHandler above, formatted by structlog)
        logging.getLogger(lgr).handlers.clear()
        logging.getLogger(lgr).propagate = True

    # Silence default uvicorn.access logs in favour of our structured alternative, attached as a middleware on the app.
    logging.getLogger("uvicorn.access").handlers.clear()



import aiofiles
import aiofiles.os
import pathlib

from .range import validate_interval

__all__ = ["stream_file"]


# TODO: support multipart/byterange responses somehow (generator of generators?)


async def stream_file(
    path: pathlib.Path,
    interval: tuple[int, int] | None,
    chunk_size: int,
    yield_content_length_as_first_8: bool = False,
    file_size: int | None = None,
    refget_mode: bool = False,
):
    """
    Stream the contents of a file, optionally yielding the content length as the first 8 bytes of the stream.
    Coordinate parameters are 0-based and inclusive, e.g., 0-10 yields the first 11 bytes. This matches the format of
    HTTP range headers.
    :param path: The path to the file to stream from.
    :param interval: Inclusive, 0-based byte interval to stream. If None, the whole file is streamed instead.
    :param chunk_size: The maximum number of bytes to read/yield at a time while streaming the file.
    :param yield_content_length_as_first_8: Whether to yield the response size as the first byte chunk (8 bytes,
           big-endian encoded) of the stream.
    :param file_size: The whole file's size, if already known. If this has already been calculated/stored, this saves a
           stat() call.
    :param refget_mode: Whether to raise Refget-compliant errors, rather than correct errors (GA4GH...), i.e.,
           400 instead of 416 for past-EOF errors.
    """

    final_file_size: int = file_size or (await aiofiles.os.stat(path)).st_size
    final_interval = interval if interval else (0, final_file_size - 1)

    # Strictly enforce interval order, since we currently can't handle inverted intervals.
    validate_interval(final_interval, final_file_size, refget_mode=refget_mode, enforce_not_inverted=True)

    start, end = final_interval
    response_size: int = end - start + 1  # Inclusive interval - need to add 1

    if yield_content_length_as_first_8:
        yield response_size.to_bytes(8, "big")

    async with aiofiles.open(path, "rb") as ff:
        # First, skip over <start> bytes to get to the beginning of the range
        await ff.seek(start)

        byte_offset: int = start
        while True:
            # Add a 1 to the amount to read if it's below chunk size, because the last coordinate is inclusive.
            data = await ff.read(min(chunk_size, end + 1 - byte_offset))
            byte_offset += len(data)
            yield data

            # If we've hit the end of the file and are reading empty byte strings, or we've reached the
            # end of our range (inclusive), then escape the loop.
            # This is guaranteed to terminate with a finite-sized file.
            if not data or byte_offset > end:
                break



import asyncio

import aiofiles
import asyncpg
import contextlib
from pathlib import Path
from typing import AsyncIterator

__all__ = [
    "PgAsyncDatabaseException",
    "PgAsyncDatabase",
]


class PgAsyncDatabaseException(Exception):
    pass


class PgAsyncDatabase:

    def __init__(self, db_uri: str, schema_path: Path):
        self._db_uri: str = db_uri
        self._schema_path: Path = schema_path

        self._pool: asyncpg.Pool | None = None
        self._pool_init_task: asyncio.Task | None = None
        self._pool_closing_task: asyncio.Task | None = None

    async def initialize(self, pool_size: int = 10) -> bool:
        if self._pool_closing_task:
            raise PgAsyncDatabaseException("Cannot open the pool while it is closing")

        if not self._pool:  # Initialize the connection pool if needed
            if not self._pool_init_task:
                async def _init():
                    pool = await asyncpg.create_pool(self._db_uri, min_size=pool_size, max_size=pool_size)

                    # If we freshly initialized the connection pool, connect to the database and execute the schema
                    # script. Don't use our own self.connect() method, since that'll end up in a circular task await.
                    async with aiofiles.open(self._schema_path, "r") as sf:
                        conn: asyncpg.Connection
                        async with pool.acquire() as conn:
                            async with conn.transaction():
                                await conn.execute(await sf.read())

                    self._pool = pool
                    self._pool_init_task = None

                    return True  # Freshly initialized the pool + executed the schema script

                self._pool_init_task = asyncio.create_task(_init())

            # self._pool_init_task is now guaranteed to not be None - can be awaited
            return await self._pool_init_task

        return False  # Pool already initialized

    async def close(self) -> bool:
        if self._pool_init_task:
            raise PgAsyncDatabaseException("Cannot close the pool while it is opening")

        if self._pool:
            if not self._pool_closing_task:
                async def _close():
                    await self._pool.close()
                    # must come after the "await" in this function, so that we can properly re-use the task that is
                    # checked for IF self._pool is not None:
                    self._pool = None
                    self._pool_closing_task = None
                    return True  # Just closed the pool

                self._pool_closing_task = asyncio.create_task(_close())

            # self._pool_closing_task is now guaranteed to not be None - can be awaited
            return await self._pool_closing_task

        return False  # Pool already closed

    @contextlib.asynccontextmanager
    async def connect(self, existing_conn: asyncpg.Connection | None = None) -> AsyncIterator[asyncpg.Connection]:
        # TODO: raise raise DatabaseError("Pool is not available") when FastAPI has lifespan dependencies
        #  + manage pool lifespan in lifespan fn.

        # If we're currently closing, wait for closing to finish before trying to re-open
        if self._pool_closing_task:
            await self._pool_closing_task

        if self._pool is None:
            # initialize if this is the first time we're using the pool, or wait for existing initialization to finish:
            await self.initialize()

        if existing_conn is not None:
            yield existing_conn
            return

        conn: asyncpg.Connection
        async with self._pool.acquire() as conn:
            yield conn
